/*--------------------------------------------------------------------------------------------------------------*/
/* 							Solves a distributed NTF problem in C++				    							*/
/*    (The implementation uses MPI for distributed communication and the Eigen Library for Linear Algebra)   	*/
/*                                                                           									*/
/* 1 - Create Data: make create_data																			*/
/*              	./create_data	                                             								*/
/*                                                                           									*/
/* 2 - Compile: make							 																*/
/*                                                                           									*/
/* 3 - Execute: mpirun -np #cores ntf			 	                    		          						*/
/*																												*/
/*																												*/
/* A. P. Liavas																									*/
/* Georgios Lourakis																							*/ 
/* Georgios Kostoulas																							*/
/* 15/11/2017                                              														*/
/*--------------------------------------------------------------------------------------------------------------*/
#include <iostream>
#include <fstream>
#include <Eigen/Dense>
#include <mpi.h>
#include <time.h> 
#include <math.h>
#include <string>
#include "ntf_functions.h"
#include <limits>

using namespace std;
using namespace Eigen;

/* <----------------------------------------------------------------------------------------------------------> */
/* < ---------------------------------------		MAIN		----------------------------------------------> */
int main(int argc, char **argv){
	int I, J, K;										// Tensor Dimensions
	int R;												// Rank of factorization
	
	const double AO_tol = 1e-3;							// Tolerance for AO Algorithm
	const double delta_1 = 1e-2;						// | Tolerance for Nesterov Algorithm
	const double delta_2 = 1e-2;						// |
	int max_iter = 500;									// Maximum Number of iterations
	
	int my_world_rank,my_rank;							// | The ids of each processor
	int frontal_rank, horizontal_rank, lateral_rank;	// |
	int column_rank, row_rank, tube_rank;				// |
	int comm_sz;										// Total number of processors
	int p_A, p_B, p_C;									// Number of processor per dimension
	int coordinates[3];
	
	double frob_X, frob_X_local;						// Frobenius norm of Tensor/Subtensors
	double f_value;										// Objective Value
	
	int acc_coeff = 3;									// | Variables 
	int acc_fail = 0;									// | for 
	int k_0 = 1;										// | Acceleration step
	
	double t_end_max, t_end_min;						// |
	double t_start, t_end, total_t;						// |
	clock_t start_t, end_t ;							// | Time variables
	int AO_iter;										// |
	
	Set_Info(&R, &I, &J, &K, "Data_cpp/info.bin");			// Initialize tensor size and rank from file
	
	MPI_Init(&argc, &argv);  								// | Initialize the MPI execution environment
	MPI_Comm_rank(MPI_COMM_WORLD, &my_world_rank); 			// | Determine current running process
	MPI_Comm_size(MPI_COMM_WORLD, &comm_sz); 				// | Total number of processes

	//	<-----------------------------		Create Communication Grid		------------------------------------>	//
	MPI_Comm grid_comm, frontal_comm, horizontal_comm, lateral_comm, column_comm, row_comm, tube_comm;
	
	Compute_Grid_Dimensions(&p_A, &p_B, &p_C, comm_sz, 1);

	Create_Grid(&grid_comm, &frontal_comm, &horizontal_comm, &lateral_comm, &column_comm, &row_comm, &tube_comm, &my_rank, &frontal_rank, &horizontal_rank, &lateral_rank, &column_rank, &row_rank, &tube_rank, coordinates, p_A, p_B, p_C);

	if (my_world_rank==0)
		cout << "R=" << R << ", I=" << I << ", J=" << J << ", K=" << K << ", comm_sz=" << comm_sz <<  ", p_A=" << p_A <<  ", p_B=" << p_B <<  ", p_C=" << p_C << endl;
	//	<------------------------------------------------------------------------------------------------------>	//
	int n_I[p_A], n_J[p_B], n_K[p_C];
	int dis_I_local[p_B*p_C], dis_J_local[p_A*p_C], dis_K_local[p_A*p_B];
	int Count_I_local[p_B*p_C], Count_J_local[p_A*p_C], Count_K_local[p_A*p_B];	
	int skip_I[p_A], skip_J[p_B], skip_K[p_C];	
	
	//	<-----------------------		Define arrays for Subtensors Dimensions		---------------------------->	//	
	Dis_Count(skip_I, n_I, p_A, I, 1);
	Dis_Count(skip_J, n_J, p_B, J, 1);
	Dis_Count(skip_K, n_K, p_C, K, 1);
	
	//	<-----------------------		Define arrays for the rows each processor updates		--------------->	//	
	Dis_Count(dis_I_local, Count_I_local, p_B*p_C, n_I[column_rank], R);
	Dis_Count(dis_J_local, Count_J_local, p_A*p_C, n_J[row_rank], R);
	Dis_Count(dis_K_local, Count_K_local, p_A*p_B, n_K[tube_rank], R);	
	
	//	<----------------------------		Matrix Initializations		--------------------------------------->	//	
	MatrixXd A(n_I[column_rank],R), B(n_J[row_rank],R), C(n_K[tube_rank],R);					// Factors A, B, C
	MatrixXd A_N(n_I[column_rank],R), B_N(n_J[row_rank],R), C_N(n_K[tube_rank],R);				// Normalized Factors A_N, B_N, C_N
	MatrixXd A_old_N(n_I[column_rank],R), B_old_N(n_J[row_rank],R), C_old_N(n_K[tube_rank],R);	// Normalized Factors A_old_N, B_old_N, C_old_N
	MatrixXd X_A(n_I[column_rank], n_J[row_rank]*n_K[tube_rank]);								// |
	MatrixXd X_B(n_J[row_rank]*n_K[tube_rank], n_I[column_rank]);								// |Matricized Tensors
	MatrixXd X_C(n_K[tube_rank], n_I[column_rank]*n_J[row_rank]);								// |
	MatrixXd A_T_A(R, R), B_T_B(R, R), C_T_C(R, R);												// A^T*A, B^T*B, C^T*C
	MatrixXd W_C(n_K[tube_rank], R);															// MTKRP for the computation of cost function
	MatrixXd KhatriRao_CB(n_J[row_rank]*n_K[tube_rank], R);										// |
	MatrixXd KhatriRao_CA(n_I[column_rank]*n_K[tube_rank], R);									// |Khatri Rao products
	MatrixXd KhatriRao_BA(n_I[column_rank]*n_J[row_rank], R);									// |
	
	//	<----------------------		Read Initial Factors and Tensor from file		--------------------------->	//
	Read_Data(A, B, C, X_A, X_B, X_C, I, J, K, n_I[column_rank], n_J[row_rank], n_K[tube_rank], skip_I[column_rank], skip_J[row_rank], skip_K[tube_rank], R, comm_sz);
 
	//	<-----------------------		Frobenious Squared Norm of Tensor		------------------------------->	//	
	frob_X_local = X_C.squaredNorm();
	MPI_Allreduce(&frob_X_local, &frob_X, 1, MPI_DOUBLE, MPI_SUM, grid_comm);			// Communication through all processes
	
	//	<-----------------------	Normalize and Communicate A^T*A, B^T*B, C^T*C		-------------------------->	//
	Normalize_Init(A, B, C, row_comm, tube_comm);

	A_T_A.noalias() = A.transpose() * A;
	MPI_Allreduce(MPI_IN_PLACE, A_T_A.data(), R*R, MPI_DOUBLE, MPI_SUM, column_comm);	// Communication through columns

	B_T_B.noalias() = B.transpose() * B;
	MPI_Allreduce(MPI_IN_PLACE, B_T_B.data(), R*R, MPI_DOUBLE, MPI_SUM, row_comm);		// Communication through rows

	C_T_C.noalias() = C.transpose() * C;
	MPI_Allreduce(MPI_IN_PLACE, C_T_C.data(), R*R, MPI_DOUBLE, MPI_SUM, tube_comm);		// Communication through tubes
	
		
	if (my_rank==0)
		cout << " BEGIN ALGORITHM " << endl;
	//	<-----------------------	Start Timers		--------------------------->	//
	MPI_Barrier(grid_comm);
	t_start = MPI_Wtime();
	start_t = clock();
	
	AO_iter = 0;
	while(1){	//	<-------------------------------------------	Begin AO Algorithm		------------------------------------------>	//
		AO_iter++;
		
		//if (my_rank==0)
		//	cout << AO_iter << " -- " << f_value/sqrt(frob_X) << " -- " << frob_X_local << " -- " << frob_X << endl;
		//	<------------------------		 Update A		------------------------------->	//
		Khatri_Rao_Product(C, B, KhatriRao_CB);															// C kr B	
		Workerjob(horizontal_rank, n_J[row_rank], n_K[tube_rank], A, C_T_C, B_T_B, X_A, KhatriRao_CB, Count_I_local, dis_I_local, delta_1, delta_2, 1, horizontal_comm);
		A_T_A.noalias() = A.transpose() * A;															// Communicate A^T*A
		MPI_Allreduce(MPI_IN_PLACE, A_T_A.data(), R*R, MPI_DOUBLE, MPI_SUM, column_comm);				// Communication through columns

		//	<------------------------		 Update B		------------------------------->	//
		Khatri_Rao_Product(C, A, KhatriRao_CA);															// C kr A		
		Workerjob(lateral_rank, n_I[column_rank], n_K[tube_rank], B, C_T_C, A_T_A, X_B, KhatriRao_CA, Count_J_local, dis_J_local, delta_1, delta_2, 2, lateral_comm);
		B_T_B.noalias() = B.transpose() * B;															// Communicate B^T*B
		MPI_Allreduce(MPI_IN_PLACE, B_T_B.data(), R*R, MPI_DOUBLE, MPI_SUM, row_comm);					// Communication through rows

		//	<------------------------		 Update C		------------------------------->	//
		Khatri_Rao_Product(B, A, KhatriRao_BA);															// B kr A
		Workerjob_C(frontal_rank, C, W_C, B_T_B, A_T_A, X_C, KhatriRao_BA, Count_K_local, dis_K_local, delta_1, delta_2, frontal_comm);
		C_T_C.noalias() = C.transpose() * C;															// Communicate C^T*C
		MPI_Allreduce(MPI_IN_PLACE, C_T_C.data(), R*R, MPI_DOUBLE, MPI_SUM, tube_comm);					// Communication through tubes
		
		//	<-----------------------	Cost Function Computation		---------------------->	//
		f_value = Get_Objective_Value(C, W_C, A_T_A, B_T_B, C_T_C, frob_X, grid_comm);

		//	<-----------------------	Normalization step		--------------------------->	//
		Normalize(A, B, C, A_T_A, B_T_B, C_T_C);
		A_N = A;
		B_N = B;
		C_N = C;
									
		//	<-----------------------	Terminating condition		----------------------->	//
		if (f_value/sqrt(frob_X) < AO_tol || AO_iter > max_iter)
			break;

		//	<-----------------------	Acceleration step		--------------------------->	//
		if (AO_iter>k_0)
			Line_Search_Accel(A_old_N, B_old_N, C_old_N, A, B, C, A_T_A, B_T_B, C_T_C, KhatriRao_BA, X_C, &acc_fail, &acc_coeff, AO_iter, f_value, frob_X, column_comm, row_comm, tube_comm, grid_comm);
			
		A_old_N = A_N;
		B_old_N = B_N;
		C_old_N = C_N;
	}	//	<-------------------------------------------	End of AO Algorithm		------------------------------------------->	//

	//	<-----------------------	End of timers		--------------------------->	//
	t_end = MPI_Wtime();
	end_t = clock();
	total_t = (end_t - start_t);

	t_end -= t_start;
	MPI_Allreduce(&t_end, &t_end_max, 1, MPI_DOUBLE, MPI_MAX, grid_comm);					// Communication through all processes	
	MPI_Allreduce(&t_end, &t_end_min, 1, MPI_DOUBLE, MPI_MIN, grid_comm);					// Communication through all processes		
	cout.precision(15);
	
	//	<----------------	Processor 0 prints some	results in the terminal	------------------------->	//
	if (my_rank == 0) {
		cout << " MPI_Wtime_max = " << t_end_max << endl;	
		cout << " MPI_Wtime_min = " << t_end_min << endl;
		cout << " CPU time = " << ((float)total_t)/CLOCKS_PER_SEC << endl;
		cout << " AO_iter = " << AO_iter << endl;
		cout << " relative f_value = " << f_value/sqrt(frob_X) << endl << endl;
	}
	//	<----------------	Processor 0 writes some results in Results.txt	------------------------->	//
/*	if (my_rank == 0) {
		ofstream my_file("Data_cpp/Results.txt", ios::out | ios::app);
		if (my_file.is_open()){
			my_file << comm_sz << " " << I << " " << J << " " << K << " " << R << " " << t_end_max << " " << ((float)total_t)/CLOCKS_PER_SEC << " " << AO_iter << " " << (f_value/sqrt(frob_X)) << " " << p_A << " " << p_B << " " << p_C;
			my_file << endl;
			my_file.close();
		}
		else
			cout << "Unable to open file";
	}
/*
	if (my_rank==0)
		cout << "Machine Epsilon is: " << numeric_limits<double>::epsilon() << endl;
*/	
	MPI_Comm_free( &tube_comm );						// |
	MPI_Comm_free( &row_comm );							// |
	MPI_Comm_free( &column_comm );						// |
	MPI_Comm_free( &lateral_comm );						// | Free the new communicators
	MPI_Comm_free( &horizontal_comm );					// |
	MPI_Comm_free( &frontal_comm );						// |
	MPI_Comm_free( &grid_comm );						// |
	
	MPI_Finalize();										// Shut down the MPI execution environment 

	return 0;	
}														// End of main
